;;; -*- mode: lisp; coding: utf-8 -*-
;;; 
;;; Copyright (c) 2007 Masayuki Onjo <onjo@lispuser.net>
;;; 
;;; Redistribution and use in source and binary forms, with or without
;;; modification, are permitted provided that the following conditions
;;; are met:
;;; 
;;;   * Redistributions of source code must retain the above copyright
;;;     notice, this list of conditions and the following disclaimer.
;;; 
;;;   * Redistributions in binary form must reproduce the above
;;;     copyright notice, this list of conditions and the following
;;;     disclaimer in the documentation and/or other materials
;;;     provided with the distribution.
;;; 
;;; THIS SOFTWARE IS PROVIDED BY THE AUTHOR 'AS IS' AND ANY EXPRESSED
;;; OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
;;; WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
;;; ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY
;;; DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
;;; DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
;;; GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
;;; INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
;;; WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
;;; NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
;;; SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

(defpackage :cdb (:use :cl) (:export ;; cdb make api
			             #:with-cdb-make
				     #:cdb-make-add
				     ;; cdb find api
				     #:with-cdb-find
				     #:cdb-findstart
				     #:cdb-findnext
				     #:cdb-get-data
				     ;; cdb utilities
				     #:cdb-get
				     #:cdb-make
				     #:cdb-dump
				     ))
(in-package :cdb)

;;; cdb default encoding (mainly for japanese user)

(defparameter *default-encoding* :utf-8)

;;; basic I/O macro

(defmacro read-long (stream)
  `(let (($stream ,stream)
	 (buffer  (make-array 4 :initial-element 0 :element-type '(unsigned-byte 8))))
     (declare (dynamic-extent buffer))
     (read-sequence buffer $stream)
     (reduce (lambda (i v)
	       (+ (ash v 8) i))
	     buffer
	     :from-end t)))

(defmacro write-long (long stream)
  `(let (($stream ,stream)
	 ($long   ,long)
	 (buffer  (make-array 4 :element-type '(unsigned-byte 8))))
     (declare (dynamic-extent buffer))
     (setf (aref buffer 0) (logand $long #xFF)
	   (aref buffer 1) (logand (ash $long -8) #xFF)
	   (aref buffer 2) (logand (ash $long -16) #xFF)
	   (aref buffer 3) (logand (ash $long -24)  #xFF))
     (write-sequence buffer $stream)))
	 
(defmacro read-long (stream)
  `(let (($stream ,stream)
	 (buffer  (make-array 4 :initial-element 0 :element-type '(unsigned-byte 8))))
     (declare (dynamic-extent buffer))
     (read-sequence buffer $stream)
     (reduce (lambda (i v)
	       (+ (ash v 8) i))
	     buffer
	     :from-end t)))

;;; cdb hash function

(defun cdb-hash (key)
  "CDB hash function"
  (reduce (lambda (h c)
	   (logxor (logand (+(logand (ash h 5) #xFFFFFFFF) h) #xFFFFFFFF) c))
	  key
	  :initial-value 5381))

;;; cdb structure

(defstruct cdb-make
  (stream      nil)
  (base        0)
  (offset      0)
  (hash-offset 0)
  (buckets     (make-array 256 :initial-element nil)))

(defstruct cdb
  (auto-close nil)
  (stream nil)
  (base   0)
  (start  0)
  (slots  0)
  (count  0)
  (data-position 0)
  (data-size 0))

;;; cdb make api

(defun cdb-make-start (stream &optional (base 0))
  (skip-cdb-header stream)
  (make-cdb-make :stream stream
		 :base   base
		 :offset (file-position stream)))

(defun cdb-make-add (cdb key value)
  (let ((key-octets   (if (stringp key) (jp:encode key *default-encoding*) key))
	(value-octets (if (stringp value) (jp:encode value *default-encoding*) value))
	(stream       (cdb-make-stream cdb)))
    ;; write one record
    (write-long (length key-octets) stream)
    (write-long (length value-octets) stream)
    (write-sequence key-octets stream)
    (write-sequence value-octets stream)
    ;; make bucket
    (let* ((hash  (cdb-hash key-octets))
	   (index (logand hash #xFF)))
      (setf (aref (cdb-make-buckets cdb) index) (nconc (aref (cdb-make-buckets cdb) index) (list (cons hash (cdb-make-offset cdb)))))
      (incf (cdb-make-offset cdb) (+ 4 4 (length key) (length value-octets))))))

(defun cdb-make-finish (cdb &aux (stream (cdb-make-stream cdb)))
  ;; write hash table
  (setf (cdb-make-hash-offset cdb) (cdb-make-offset cdb))
  (loop for bucket across (cdb-make-buckets cdb)
        do
        (when bucket
	  (let* ((num   (* (length bucket) 2))
		 (cells (make-array num :initial-element (cons 0 0))))
	    (loop with i = 0
	       for (hash . pos) in bucket
	       do
	       (setf i (mod (ash hash -8) num))
	       (loop while (/= (cdr (aref cells i)) 0) do (setf i (mod (1+ i) num)))
	       (setf (aref cells i) (cons hash pos)))
	    (loop for (hash . pos) across cells do
		 (write-long hash stream)
		 (write-long pos stream)))))
    ;; write pointers
    (file-position stream (cdb-make-base cdb))
    (loop for bucket across (cdb-make-buckets cdb)
	  for num = (length bucket)
	  for i from 0
	  do (write-long (cdb-make-hash-offset cdb) stream)
	     (write-long (* num 2) stream)
	     (incf (cdb-make-hash-offset cdb) (* num 2 (+ 4 4)))))

(defmacro with-cdb-make ((cdb file) &body body)
  (let (($stream (gensym)))
    `(with-open-file (,$stream ,file :direction :output :if-exists :supersede :element-type '(unsigned-byte 8))
       (let ((,cdb (cdb-make-start ,$stream)))
	 (when ,cdb
	   (unwind-protect
		(progn ,@body)
	     (cdb-make-finish ,cdb)))))))

;;; cdb find api

(defun cdb-init (stream-or-filename &optional (cdb (make-cdb)))
  (let* ((auto-close nil)
	 (stream (typecase stream-or-filename
		   ((or string pathname)
		    (setf auto-close T)
		    (open stream-or-filename :direction :input :element-type '(unsigned-byte 8) #+sbcl :mapped #+sbcl T))
		   (stream
		    stream-or-filename)
		   (T
		    (error "cdb-init parameter error: parameter ~A" stream-or-filename)))))
    (setf (cdb-auto-close cdb) auto-close
	  (cdb-stream cdb) stream
	  (cdb-count  cdb) 0)
    cdb))

(defun cdb-free (cdb)
  (typecase cdb
    (cdb
     (when (cdb-auto-close cdb)
       (close (cdb-stream cdb))
       (setf (cdb-stream cdb) nil))
     (setf (cdb-count cdb) 0))
    (T
     (error "cdb-free parameter error: parameter ~A" cdb))))

(defun cdb-findstart (cdb)
  (setf (cdb-data-position cdb) 0
	(cdb-data-size cdb) 0
	(cdb-count cdb) 0))

(defun cdb-findnext (cdb key)
  (let* ((key-octets (if (stringp key) (jp:encode key *default-encoding*) key))
	 (hash (cdb-hash key-octets))
	 (index (logand hash #xFF))
	 (base (cdb-base cdb))
	 (stream (cdb-stream cdb))
	 (bucket 0)
	 (entries 0))
    ;; first time only
    (when (= (cdb-count cdb) 0)
      ;; p0 p1 ... p255 : N = (hash & 0xFF)
      (file-position stream (+ base (* index (+ 4 4))))
      (setf bucket  (read-long stream)
	    entries (read-long stream)
	    (cdb-slots cdb) entries
	    (cdb-start cdb) (+ base bucket (* (mod (ash hash -8) entries) (+ 4 4))))
      ;; no entry
      (when (= entries 0)
	(return-from cdb-findnext nil)))
    ;; lookup hash0 ... hash255
    (loop while (< (cdb-count cdb) (cdb-slots cdb))
          do
	  (file-position stream (cdb-start cdb))
	  (incf (cdb-count cdb) 1)
	  (incf (cdb-start cdb) (+ 4 4))
	  (let* ((code (read-long stream))
		 (pos  (read-long stream)))
	    ;; compare hash code
	    (when (= code hash)
	      (file-position stream (+ base pos))
	      (let* ((klen (read-long stream))
		     (vlen (read-long stream))
		     (k    (make-array klen :element-type '(unsigned-byte 8))))
		(read-sequence k stream)
		;; compare key
		(when (loop for e1 across k
			    for e2 across key-octets
			    always (= e1 e2))
		  (setf (cdb-data-position cdb) (file-position stream)
			(cdb-data-size cdb)     vlen)
		  (return-from cdb-findnext T))))))))

(defun cdb-get-data (cdb &key (binary nil))
  (when (and (/= (cdb-data-position cdb) 0)
	     (file-position (cdb-stream cdb) (cdb-data-position cdb)))
    (let ((vec (make-array (cdb-data-size cdb) :element-type '(unsigned-byte 8))))
      (read-sequence vec (cdb-stream cdb))
      (if binary vec (jp:decode vec *default-encoding*)))))

(defmacro with-cdb-find ((cdb file-or-stream) &body body)
  `(let ((,cdb (cdb-init ,file-or-stream)))
     (unwind-protect
	  (progn ,@body)
       (cdb-free ,cdb))))

;;; cdb utilities  
    
(defun cdb-get (file-or-stream key &key (offset 1) (binary nil))
  (with-cdb-find (cdb file-or-stream)
    (cdb-findstart cdb)
    (loop for i from 0 below offset
          while (cdb-findnext cdb key)
	  finally
	 (when (= i offset)
	   (return (cdb-get-data cdb :binary binary))))))

(defun cdb-make (file-or-stream hash-table)
  (if (streamp file-or-stream)
      (%cdb-make file-or-stream (file-position file-or-stream) hash-table)
      (with-open-file (stream file-or-stream :direction :output :if-exists :supersede :element-type '(unsigned-byte 8))
	(%cdb-make stream (file-position stream) hash-table))))

(defun skip-cdb-header (stream &optional (base 0))
  (declare (ignorable base))
  #-clisp (file-position stream (+ base (* 256 (+ 4 4))))
  #+clisp (write-sequence (make-array (* 256 (+ 4 4)) :initial-element 0) stream))

(defun %cdb-make (stream base hash-table)
  (let ((buckets     (make-array 256 :initial-element nil))
	(hash-offset 0))
    ;; skip pointers
    (skip-cdb-header stream)
    ;; write records
    (loop with offset = (file-position stream)
          for k being each hash-key of hash-table using (hash-value v)
	  for key = (if (stringp k) (jp:encode k *default-encoding*) k)
	  for val = (if (stringp v) (jp:encode v *default-encoding*) v)
	  do
	  ;; write one record
	  (write-long (length key) stream)
	  (write-long (length val) stream)
	  (write-sequence key stream)
	  (write-sequence val stream)
	  ;; make bucket
	  (let* ((hash  (cdb-hash key))
		 (index (logand hash #xFF)))
	    (setf (aref buckets index) (nconc (aref buckets index) (list (cons hash offset))))
	    (incf offset (+ 4 4 (length key) (length val))))
	 finally (setf hash-offset offset))
    ;; write hash table
    (loop for bucket across buckets
	  do
	  (when bucket
	    (let* ((num   (* (length bucket) 2))
		   (cells (make-array num :initial-element (cons 0 0))))
	      (loop with i = 0
		    for (hash . pos) in bucket
		    do
		    (setf i (mod (ash hash -8) num))
		    (loop while (/= (cdr (aref cells i)) 0) do (setf i (mod (1+ i) num)))
		    (setf (aref cells i) (cons hash pos)))
	      (loop for (hash . pos) across cells do
		    (write-long hash stream)
		    (write-long pos stream)))))
    ;; write pointers
    (file-position stream base)
    (loop for bucket across buckets
	  for num = (length bucket)
	  for i from 0
	  do (write-long hash-offset stream)
	     (write-long (* num 2) stream)
	     (incf hash-offset (* num 2 (+ 4 4))))))

(defun test ()
  (let ((hash (make-hash-table :test #'equal)))
    (setf (gethash "one" hash) "Hello"
	  (gethash "two" hash) "Goodbye"
	  (gethash "three" hash) "Good Morning")
    (cdb-make "test.cdb" hash)
    (cdb-get "test.cdb" "one")))

(defun test2 ()
  (with-cdb-make (cdb "test2.cdb")
    (cdb-make-add cdb "one" "Hello1")
    (cdb-make-add cdb "one" "Hello2")
    (cdb-make-add cdb "one" "Hello3")
    (cdb-make-add cdb "one" "Hello4"))
  (let ((cdb (cdb-init "test2.cdb")))
    (unwind-protect
	 (progn
	   (cdb-findstart cdb)
	   (loop while (cdb-findnext cdb "one")
	         do (print (cdb-get-data cdb))))
      (cdb-free cdb))))
    

(defun test3 (&aux search)
  (flet ((make-random-record (cdb)
	   (let* ((s1 (1+ (random 10)))
		  (s2 (1+ (random 100)))
		  (key (make-array s1 :element-type '(unsigned-byte 8) :initial-element 0))
		  (value (make-array s2 :element-type '(unsigned-byte 8) :initial-element 0)))
	     (loop for i from 0 below s1 do (setf (aref key i) (logand (+ (char-code #\A) (random 26)) #xFF)))
	     (loop for i from 0 below s2 do (setf (aref value i) (logand (+ (char-code #\A) (random 52)) #xFF)))
	     (cdb-make-add cdb key value)
	     (setf search (cons key value)))))
    (time
     (with-cdb-make (cdb "test3.cdb")
       (loop repeat 100000 do (make-random-record cdb))))
    (time
     (let ((result (cdb-get "test3.cdb" (car search) :binary t)))
       (format t "~S~%" search)
       (loop for a across result
	     for b across (cdr search)
	     always (= a b))))))

#||
CDB FORMAT <http://cr.yp.to/cdb/cdb.txt>

A structure for constant databases
19960914
Copyright 1996
D. J. Bernstein, djb@pobox.com

A cdb is an associative array: it maps strings (``keys'') to strings
(``data'').

A cdb contains 256 pointers to linearly probed open hash tables. The
hash tables contain pointers to (key,data) pairs. A cdb is stored in
a single file on disk:

    +----------------+---------+-------+-------+-----+---------+
    | p0 p1 ... p255 | records | hash0 | hash1 | ... | hash255 |
    +----------------+---------+-------+-------+-----+---------+

Each of the 256 initial pointers states a position and a length. The
position is the starting byte position of the hash table. The length
is the number of slots in the hash table.

Records are stored sequentially, without special alignment. A record
states a key length, a data length, the key, and the data.

Each hash table slot states a hash value and a byte position. If the
byte position is 0, the slot is empty. Otherwise, the slot points to
a record whose key has that hash value.

Positions, lengths, and hash values are 32-bit quantities, stored in
little-endian form in 4 bytes. Thus a cdb must fit into 4 gigabytes.

A record is located as follows. Compute the hash value of the key in
the record. The hash value modulo 256 is the number of a hash table.
The hash value divided by 256, modulo the length of that table, is a
slot number. Probe that slot, the next higher slot, and so on, until
you find the record or run into an empty slot.

The cdb hash function is ``h = ((h << 5) + h) ^ c'', with a starting
hash of 5381.
||#
